import math

import pandas
from loguru import logger
try:
	from muller.inheritance.areascore import area_of_series, calculate_common_area
except ModuleNotFoundError:
	try:
		from ...inheritance.areascore import area_of_series, calculate_common_area
	except ValueError:
		from muller.inheritance.areascore import area_of_series, calculate_common_area


def minkowski_distance(left: pandas.Series, right: pandas.Series, p: int = 2) -> float:
	""" Calculates the minkowski distance between two series. Essentially just a generic lp-norm.
		Parameters
		----------
		left:pandas.Series
		right:pandas.Series
		p:int
	"""
	total = sum([math.pow(abs(i - j), p) for i, j in zip(left.tolist(), right.tolist())])
	return math.pow(total, 1 / p)


def pearson_correlation_distance(left: pandas.Series, right: pandas.Series, adjusted = True) -> float:
	"""
		Calculates the pearson correlation between two series. The resulting value lies in the range [0,2].
	Parameters
	----------
	left:pandas.Series
	right:pandas.Series
	adjusted: bool; default True
		Whther to adjust the coefficient for small sample sizes.

	Returns
	-------
	float
	"""

	pcc = left.corr(right, method = 'pearson')
	# Adjust due to sample size
	if adjusted:
		adjusted_pcc = adjust_correlation_coefficient(pcc, len(left))
	else:
		adjusted_pcc = pcc
	# convert to distance metric.
	return 1 - adjusted_pcc


def adjust_correlation_coefficient(r: float, n: int) -> float:
	value = (1 - r ** 2) / (2 * n)
	ra = r * (1 + value)
	return ra


# noinspection PyTypeChecker
def binomial_distance(left: pandas.Series, right: pandas.Series) -> float:
	""" Based on the binomial calculations present in the original matlab scripts."""
	# Find the mean frequency of each timepoint
	# index is timepoints,  values are frequencies
	not_detected_fixed_df = pandas.concat([left, right], axis = 1)
	mean: pandas.Series = not_detected_fixed_df.mean(axis = 1)

	# Calculate sigma_freq
	# E(sigma) = (1/n) sum(sigma) = (1/n) sum(np(1-p)) == sum(p(1-p)
	# E(sigma_p) = (1/n) E(sigma) == 1/n(sum(p(1-p))
	# E(d_bar) = 1/n(sum(di)) == 1/n (n*sum(di))
	# pandas.Series.radd is slow for some reason. Use '-' operator instead.
	n = len(mean)
	sigma_freq: pandas.Series = mean.mul(1 - mean)
	# Difference of frequencies at each timepoint
	# difference: pandas.Series = not_detected_fixed_df.iloc[:, 0] - not_detected_fixed_df.iloc[:, 1]
	difference = not_detected_fixed_df.diff(axis = 1).iloc[:, 1]

	sigma_pair: float = sigma_freq.sum() / n ** 2
	# Sum of differences
	difference_mean: float = abs(difference).sum() / n

	X = difference_mean / (math.sqrt(2 * sigma_pair))

	return X


def binomial_probability(left: pandas.Series, right: pandas.Series) -> float:
	""" Calculates the probability that the distance between `left` and `right` can be explained by experimental error.
		Note that higher probabilities indicate a closer relationship between two series in contrast to distance measurements where
		lower distances correspond to better matches. This can be accounted for by subtracting the probability value from 1.

	"""
	X = binomial_distance(left, right)
	value = 1 - math.erf(X)
	logger.error("Set a patch so that binomialp represents probability")
	value = 1- value
	return value


def jaccard_distance(left: pandas.Series, right: pandas.Series) -> float:
	area_left = area_of_series(left)
	area_right = area_of_series(right)
	area_shared = calculate_common_area(left, right)
	j = area_shared / (area_left + area_right - area_shared)
	return 1 - j


def calculate_distance(left: pandas.Series, right: pandas.Series, metric: str) -> float:
	if metric == 'binomial':
		# default option so placed first to minimize equality calculations.
		distance_between_series = binomial_distance(left, right)
	elif metric == 'pearson':
		distance_between_series = pearson_correlation_distance(left, right)
	elif metric == 'minkowski':
		distance_between_series = minkowski_distance(left, right)
	elif metric == 'similarity' or metric == 'binomialp':
		distance_between_series = binomial_probability(left, right)
	elif metric == 'jaccard':
		distance_between_series = jaccard_distance(left, right)
	elif metric == "combined":
		distance_between_series_pearson = pearson_correlation_distance(left, right)
		distance_between_series_minkowski = minkowski_distance(left, right, 2)
		distance_between_series = (2 * distance_between_series_pearson) + distance_between_series_minkowski
	else:
		message = f"'{metric}' is not an available metric."
		raise ValueError(message)
	return distance_between_series



import itertools
from pathlib import Path
from typing import Dict, Iterable, List, Tuple

import pandas
from scipy.spatial import distance

PairwiseArrayType = Dict[Tuple[str, str], float]

# TODO: Refactor using UserDict
class DistanceCache:
	"""
		Calculates and holds the distances for all pairwise elements.

		Usage
		-----
		calculation = PairwiseCalculation()
		# Update values
		pair_array = calculation.update_values(timeseries, detection_cutoff, fixed_cutoff)
	"""

	def __init__(self, pairwise_array: PairwiseArrayType = None):
		if pairwise_array is None: pairwise_array = dict()
		self.pairwise_values = dict()

		for (left, right), value in pairwise_array.items():
			self.pairwise_values[left, right] = value
			self.pairwise_values[right, left] = value

	def __bool__(self) -> bool:
		return bool(self.pairwise_values)

	def __len__(self) -> int:
		return len(self.pairwise_values)
	def __getitem__(self, item):
		return self.pairwise_values[item]
	def asdict(self) -> PairwiseArrayType:
		return self.pairwise_values

	def squareform(self)->pandas.DataFrame:
		""" Converts a dictionary with all pairwise values for a set of points into a square matrix representation.
		"""
		keys = sorted(set(itertools.chain.from_iterable(self.pairwise_values.keys())))
		_square_map = dict()
		for left in keys:
			series = dict()
			for right in keys:
				value = self.get(left, right, 0)
				series[right] = value
			_square_map[left] = series
		return pandas.DataFrame(_square_map)

	def triangle(self):
		""" Returns the condensed squareform of the pair array."""
		# noinspection PyTypeChecker
		return distance.squareform(self.squareform().values)

	def get(self, left, right, default = None) -> float:
		try:
			result = self.pairwise_values[left, right]
		except KeyError:
			result = self.pairwise_values.get((right, left), default)
		return result

	def reduce(self, labels: Iterable[str]) -> 'DistanceCache':
		"""
			Removes all labels from `self.pairwise_values` that are not present in `labels`
		"""
		labels = set(labels)
		pair_array = {
			k: v for k, v in self.pairwise_values.items()
			if (k[0] in labels and k[1] in labels)
		}
		self.pairwise_values = pair_array
		return self

	def update(self, pair_array: PairwiseArrayType) -> 'DistanceCache':
		for (left, right), value in pair_array.items():
			self.pairwise_values[left, right] = value
			self.pairwise_values[right, left] = value
		return self

	def unique(self) -> List[Tuple[str, str]]:
		seen = set()
		for key in sorted(self.pairwise_values.keys()):
			if key in seen or key[::-1] in seen:
				continue
			else:
				seen.add(key)
				yield key

	def save(self, filename: Path):
		with filename.open('w') as output:
			for key in self.unique():
				value = self.get(*key)
				line = f"{key[0]}\t{key[1]}\t{value}\n"
				output.write(line)

	@property
	def values(self)->List[float]:
		return list(self.pairwise_values.values())

	@classmethod
	def read(cls, filename: Path) -> 'DistanceCache':
		contents = filename.read_text().split('\n')
		contents = [i.split('\t') for i in contents]
		data = dict()
		for line in contents:
			left, right, value = line
			value = float(value)
			data[left, right] = value
			data[right, left] = value
		return DistanceCache(data)

	@classmethod
	def from_squareform(cls, square:pandas.DataFrame)->'DistanceCache':
		data = dict()

		for left, row in square.iterrows():
			for right, value in row.items():
				data[left,right] = value
				data[right,left] = value
		return DistanceCache(data)



#!python
"""
	Main script to run the muller workflow.
	.
	|---- .clade.annotated.(png|svg)
	|---- .genotypes.tsv
	|---- .lineage.clade.(png|svg)
	|---- figures/
	|----|---- clade/
	|----|----|---- .clade.annotated.(png|svg)
	|----|----|---- .clade.unannotated.(png|svg)
	|----|----|---- .clade.genotypes.(png|svg)
	|----|----|---- .clade.trajectories.(png|svg)
	|----|---- unique/
	|----|----|---- .unique.annotated.(png|svg)
	|----|----|---- .unique.unannotated.(png|svg)
	|----|----|---- .unique.genotypes.(png|svg)
	|----|----|---- .unique.trajectories.(png|svg)
	|----|---- .dendrogram.png
	|----|---- .pairwisedistance.svg
	|---- scripts/
	|----|---- .r
	|---- suplementary-files/
	|----|---- .genotypeinformation.json
	|----|---- .lineagescores.tsv
	|----|---- .options.json
	|---- tables/
	|----|---- .trajectories.tsv
	|----|---- .genotypes.tsv
	|----|---- .distance.tsv
	|----|---- .linkagematrix
	|----|---- .ggmuller.edges.tsv
	|----|---- .ggmuller.populations.tsv
	|----|---- .mullerdataframe.tsv
"""
import argparse
from pathlib import Path
from typing import Dict, List, Optional, Union
from muller import graphics, widgets
from muller.graphics import graphicsio
import pandas
pandas.set_option('mode.chained_assignment', None) # This disables the warning about setting a value on a copy of a dataframe.
from loguru import logger

from muller import clustering, dataio, inheritance, commandline_parser
from muller.dataio import projectdata, annotations

logger.remove() # Need to remove the default sink so that the logger doesn't print messages twice.
import sys
if commandline_parser.DEBUG:
	logger.add(sys.stderr, level = "DEBUG")
else:
	logger.add(sys.stderr, level = 'INFO', format = "{time:YYYY-MM-DD HH:mm:ss} {level} {message}")

def run_genotype_inference_workflow(trajectoryio: Union[str, Path, pandas.DataFrame], metric: str, dlimit: float, slimit: float, flimit: float,
		pvalue: float, known_genotypes: Optional[Path] = None, threads: Optional[int] = None) -> projectdata.DataGenotypeInference:
	"""
	Parameters
	----------
	trajectoryio: Union[str,Path,pandas.DataFrame]
		Either the path to a table with mutational trajectories or the table itself.
		The table should already be cleaned/filtered.
		The table should be organized with timepoints as columns and rows as genotype frequencies.
		Ex.
		genotype	0	7	13 ...
		genotype-1	0	.1	.3 ...
		genotype-2	0	.2	.1 ...
	metric: Literal['']
	dlimit, slimit, flimit, pvalue: float
	known_genotypes
	threads
	"""
	if isinstance(trajectoryio, (str, Path)):
		logger.info(f"Reading '{trajectoryio}' as the trajectory table.")
		trajectories, trajectory_info = dataio.parse_trajectory_table(trajectoryio)
	else:
		trajectories = trajectoryio
		trajectory_info = None

	# Read in the starting genotypes.
	if known_genotypes:
		known_genotypes = dataio.parse_known_genotypes(known_genotypes)

	genotype_generator = clustering.ClusterMutations(
		metric = metric,
		dlimit = dlimit,
		slimit = slimit,
		flimit = flimit,
		pvalue = pvalue,
		starting_genotypes = known_genotypes,
		threads = threads
	)

	genotype_data = genotype_generator.run(trajectories)
	genotype_data.table_trajectories_info = trajectory_info

	return genotype_data


def run_genotype_lineage_workflow(genotypeio: Union[str, Path, pandas.DataFrame], dlimit: float, flimit: float, pvalue: float,
		known_ancestry: Optional[Path]) -> projectdata.DataGenotypeLineage:
	"""

	Parameters
	----------
	genotypeio
	dlimit
	flimit
	pvalue
	known_ancestry

	Returns
	-------

	"""
	lineage_generator = inheritance.LineageWorkflow(
		dlimit = dlimit,
		flimit = flimit,
		pvalue = pvalue
	)

	# Read in the input data if it is not already a pandas.DataFrame object
	if isinstance(genotypeio, (str, Path)):
		logger.info(f"Reading '{genotypeio}' as the genotype table.")
		genotypes = dataio.parse_genotype_table(genotypeio)
	else:
		genotypes = genotypeio

	lineage_data = lineage_generator.run(genotypes, known_ancestry)
	return lineage_data




def get_base_filename(filename:Path, name:Optional[str], sheetname:Optional[str]) -> str:
	if name:
		base_filename = name
	else:
		base_filename = filename.stem
	if sheetname and sheetname != 'Sheet1':
		base_filename += '.' + sheetname

	return base_filename

def run_workflow(program_options: argparse.Namespace):
	logger.info("Running with the following parameters")
	for key, value in vars(program_options).items():
		logger.info(f"\t{key:<30}{value}")
	output_folder = program_options.output_folder


	data_basic = projectdata.DataWorkflowBasic(
		version = commandline_parser.__VERSION__,
		filename = program_options.filename,
		program_options = program_options
	)
	if not data_basic.program_options.output_folder.exists():
		data_basic.program_options.output_folder.mkdir()

	trajectory_table, trajectory_info = dataio.parse_trajectory_table(program_options.filename, program_options.sheetname)

	# Need to read in the input dataset.

	result_genotype_inference = run_genotype_inference_workflow(
		trajectory_table,
		program_options.metric,
		dlimit = program_options.dlimit,
		slimit = program_options.slimit,
		flimit = program_options.flimit,
		pvalue = program_options.pvalue,
		known_genotypes = program_options.known_genotypes,
		threads = program_options.threads
	)

	genotype_annotations = annotations.read_genotype_annotations(trajectory_info, result_genotype_inference.genotype_members)

	result_genotype_lineage = run_genotype_lineage_workflow(
		result_genotype_inference.table_genotypes,  # should already be sorted.
		dlimit = program_options.dlimit,
		flimit = program_options.flimit,
		pvalue = program_options.pvalue,
		known_ancestry = program_options.known_ancestry
	)


	save_tables(data_basic, result_genotype_inference, result_genotype_lineage, genotype_annotations)
	# Save using the older graphics workflow for now.
	render_graphics(
		data_basic = data_basic,
		data_inference = result_genotype_inference,
		data_lineage = result_genotype_lineage,
		genotype_annotations = genotype_annotations
	)

	data_basic.save(output_folder)

def save_tables(data_basic: projectdata.DataWorkflowBasic, data_inference: projectdata.DataGenotypeInference,
		data_lineage: projectdata.DataGenotypeLineage, genotype_annotations:Dict[str,List[str]]):
	"""
		.figures
		|---- tables
		|----|---- genotypes.tsv
		|----|---- trajectories.tsv
		|----|---- trajectores.filtered.tsv
		|----|---- edges.tsv
		|----|---- populations.tsv
		|----|---- distances.tsv
		|----|---- scorerecords.tsv
		|----|---- linkagetable.tsv
		|----|---- parameters.json
		|----|---- genotypeinformation.json
		|---- scripts
		|----|---- lineage.r
	"""
	delimiter = "\t"
	suffix = "tsv"
	prefix = get_base_filename(data_basic.program_options.filename, data_basic.program_options.name, data_basic.program_options.sheetname)
	folder_tables = widgets.checkdir(data_basic.program_options.output_folder / "tables")
	folder_data = widgets.checkdir(data_basic.program_options.output_folder / "data")
	data_basic.save(folder_data)
	data_inference.save(folder_tables, prefix)
	data_lineage.save(folder_tables, prefix)
	"""
	filename_table_genotypes = folder_table / f"{prefix}.genotypes.{suffix}"
	filename_table_trajectories = folder_table / f"{prefix}.trajectories.{suffix}"
	filename_table_trajectories_filtered = folder_table / f"{prefix}.trajectories.filtered.{suffix}"
	filename_table_edges = folder_table / f"{prefix}.edges.{suffix}"
	filename_table_population = folder_table / f"{prefix}.population.{suffix}"
	filename_table_distances = folder_table / f"{prefix}.distances.{suffix}"
	filename_table_scorerecords = folder_table / f"{prefix}.scorerecords.{suffix}"
	filename_table_linkage = folder_table / f"{prefix}.linkage.{suffix}"

	filename_data_parameters = folder_data / f"{prefix}.parameters.json"
	filename_data_genotypes = folder_data / f"{prefix}.genotypes.json"

	data_inference.table_genotypes.to_csv(filename_table_genotypes, sep = delimiter)
	data_inference.table_trajectories.to_csv(filename_table_trajectories, sep = delimiter)
	data_inference.matrix_distance.squareform().to_csv(filename_table_distances, sep = delimiter)
	data_inference.clusterdata.table_linkage.to_csv(filename_table_linkage, sep = delimiter)
	
	data_lineage.table_edges.to_csv(filename_table_edges, sep = delimiter)
	data_lineage.table_populations.to_csv(filename_table_population, sep = delimiter)
	data_lineage.table_scores.to_csv(filename_table_scorerecords, sep = delimiter)
	"""


def render_graphics(data_basic: projectdata.DataWorkflowBasic, data_inference: projectdata.DataGenotypeInference,
		data_lineage: projectdata.DataGenotypeLineage, genotype_annotations:Dict[str,List[str]]):
	""" Graphics are parametrized by filename suffix (png vs svg) and palette.
		.figures
		|---- dendrogram.png
		|---- heatmap.png
		|---- distancedistribution.png
		|---- unique palette
		|    |---- muller.annotated.(svg|png)
		|    |---- muller.unannotated.(svg|png)
		|    |---- timeseriespanel.(svg|png)
		|    |---- lineageplot.(svg|png)
		|---- lineage palette
		|    |---- muller.annotated.(svg|png)
		|    |---- muller.unannotated.(svg|png)
		|    |---- timeseriespanel.(svg|png)
		|    |---- lineageplot.(svg|png)
	"""
	custom_palette = dataio.read_map(data_basic.program_options.genotype_palette_filename)

	workflow_graphics = graphicsio.OutputGeneratorGraphics(
		project_folder = data_basic.program_options.output_folder,
		palettes = [],
		sample_basename = get_base_filename(data_basic.program_options.filename, data_basic.program_options.name, data_basic.program_options.sheetname),
		render = True
	)
	prefix = get_base_filename(data_basic.program_options.filename, data_basic.program_options.name, data_basic.program_options.sheetname)
	folder_figures = widgets.checkdir(data_basic.program_options.output_folder)
	filename_dendrogram = folder_figures / "dendrogram.png"
	filename_heatmap = folder_figures / "heatmap.png"
	filename_distance_distribution = folder_figures / "distancedistribution.png"

	# Plot the figures that aren't parametrized
	workflow_graphics.generate_dendrogram(data_inference.clusterdata.table_linkage, data_inference.matrix_distance, filename_dendrogram)
	workflow_graphics.generate_heatmap(data_inference.matrix_distance.squareform(), filename_heatmap)
	graphics.generate_distance_plot(
		data_inference.matrix_distance.values,
		data_inference.clusterdata.similarity_cutoff,
		filename_distance_distribution
	)
	# Set up the generators
	generator_panel_timeseries = graphics.TimeseriesPanel(render = data_basic.program_options.render)
	generator_plot_timeseries = graphics.TimeseriesPlot(render = data_basic.program_options.render)
	generator_plot_muller = graphics.MullerPlot(outlines = data_basic.program_options.draw_outline, render = data_basic.program_options.render)
	generator_plot_muller_panel = graphics.MullerPanel()
	# These figures need to be parametrized by palette type and filetype.
	for palette_name in ["unique", "lineage"]:
		folder_palette = widgets.checkdir(folder_figures / f"{palette_name}")

		current_palette_data = graphics.palettes.generate_palette(
			edges = data_lineage.table_edges,
			custom_palette = custom_palette,
			annotations = genotype_annotations,
			kind = palette_name,
		)

		current_palette = graphics.Palette(
			name = palette_name,
			palette = current_palette_data,
			members = data_inference.genotype_members
		)

		for suffix in ["png", "svg"]:
			filename_mullerplot_annotated = folder_palette / f"{prefix}.muller.annotated.{suffix}"
			filename_mullerplot_unannotated = folder_palette / f"{prefix}.muller.unannotated.{suffix}"
			filename_timeseries_panel = folder_palette / f"{prefix}.timeseriespanel.{suffix}"
			filename_timeseries_genotypes = folder_palette / f"{prefix}.timeseries.genotypes.{suffix}"
			filename_lineageplot = folder_palette / f"{prefix}.lineageplot.{suffix}"
			filename_muller_panel = folder_palette / f"{prefix}.mullerpanel.{suffix}"
			filename_graphviz_script = folder_palette / f"{prefix}.lineagescript.{suffix}.dot"
			filename_lineage_panel = folder_palette / f"{prefix}.lineagepanel.{suffix}"
			logger.info("Generating the panels...")

			generator_plot_timeseries.plot(
				data_inference.table_genotypes,
				palette = current_palette.get_genotype_palette(),
				filename = filename_timeseries_genotypes
			)

			generator_panel_timeseries.run(
				timeseries_genotype = data_inference.table_genotypes,
				timeseries_trajectory = data_inference.table_trajectories,
				filename = filename_timeseries_panel,
				palette = current_palette,
			)
			logger.info("Generating the annotated muller plots...")
			generator_plot_muller.plot(
				muller_df = data_lineage.table_muller,
				color_palette = current_palette.get_genotype_palette(),
				annotations = genotype_annotations,
				filename = filename_mullerplot_annotated
			)

			generator_plot_muller_panel.plot(
				timeseries = data_inference.table_genotypes,
				muller_df = data_lineage.table_muller,
				palette = current_palette.get_genotype_palette(),
				annotations = genotype_annotations,
				filename = filename_muller_panel
			)

			logger.info("Generating the unannotated muller plots...")
			generator_plot_muller.plot(
				muller_df = data_lineage.table_muller,
				color_palette = current_palette.get_genotype_palette(),
				annotations = None,
				filename = filename_mullerplot_unannotated
			)
			logger.info("Generating the lineageplot...")
			lineageplot = graphics.flowchart(
				edges = data_lineage.table_edges.reset_index(),
				palette = current_palette.get_genotype_palette(),
				annotations = genotype_annotations,
				filename = filename_lineageplot
			)

			filename_graphviz_script.write_text(lineageplot.to_string())
			if suffix == 'png':
				# PIL doesn't work with svgs
				logger.info("Generating the lineage panel...")
				graphics.generate_lineage_panel(filename_mullerplot_annotated, filename_lineageplot, filename_lineage_panel)





from collections import OrderedDict
from typing import Dict, List
import matplotlib.colors
from . import colorset
import re

available_colors = matplotlib.colors.get_named_colors_mapping()

def generate_distinctive_palette(genotypes: List[str]) -> Dict[str, str]:
	""" Assigns a unique color to each genotype."""
	# TODO: If the genotype label specifies a color, use that instead.
	color_palette = colorset.get_distinctive_palette(len(genotypes))
	if re.search("genotype-[\d]+", genotypes[0]):
		genotype_labels = sorted(genotypes, key = lambda s: int(s.split('-')[-1]))
	else:
		genotype_labels = genotypes
	# Use an OrderedDict to help with providing the correct order for the r script.
	color_map = OrderedDict()
	for label, color in zip(genotype_labels, color_palette):
		# If the genotype name specifies a color, use that.
		genotype_name = label.split('-')[1]
		if genotype_name in available_colors:
			color_map[label] = available_colors[genotype_name]
		else:
			color_map[label] = color

	return color_map



import random
from typing import List, Tuple

import seaborn

distinctive_palette = [
	'#e6194b', '#3cb44b', '#ffe119', '#4363d8', '#f58231',
	'#911eb4', '#46f0f0', '#f032e6', '#bcf60c', '#fabebe',
	'#008080', '#e6beff', '#9a6324', '#fffac8', '#800000',
	'#aaffc3', '#808000', '#ffd8b1', '#000075', '#808080'
]

similar_colorschemes = {
	'reds':   ["Reds", "Wistia", "Reds_r", "RdPu_r", "YlOrBr", "hot", "OrRd_r", "hot", "afmhot", "YlOrRd_r"],
	'blues':  ["Blues", "Purples", "BuPu", "PuBu", "cool"],
	'greens': ["Greens", "BuGn", "YlGn", "summer"],
	'greys':  ["Greys", "binary", "bone"],
	'browns': ["pink", "copper", "gist_heat"]
}

distinctive_colorschemes = [
	'Reds', 'Blues', 'Greens', 'Purples', 'hot', 'Greys', 'copper', 'cool'
]
sequential_colorschemes = [
	'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
	'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
	'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']
sequential2_colorschemes = [
	'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
	'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
	'hot', 'afmhot', 'gist_heat', 'copper']
misc_colorschemes = [
	'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
	'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
	'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', 'gist_ncar']
# For some reason seaborn doesn't like the misc colorschemes
distinctive_colorschemes = distinctive_colorschemes + sequential_colorschemes + sequential2_colorschemes# + misc_colorschemes


def load_colorscheme(string: str, size: int) -> List[str]:
	"""Attempts to load a given colorscheme."""

	try:
		palette = seaborn.color_palette(string, size)
	except ValueError:
		palette = seaborn.light_palette(string, size)

	return palette.as_hex()


def rgbtohex(rgb: Tuple[float, float, float]) -> str:
	if isinstance(rgb[0], float):  # The values are formatted as a number between 0 and 1
		red = int(rgb[0] * 256)
		green = int(rgb[1] * 256)
		blue = int(rgb[2] * 256)
	else:
		red, green, blue = rgb
	hex_string = f"#{red:>02X}{green:>02X}{blue:>02X}"
	return hex_string


def random_color(lower: int = 50, upper: int = 230) -> str:
	r = random.randint(lower, upper)
	g = random.randint(lower, upper)
	b = random.randint(lower, upper)
	color = "#{:>02X}{:>02X}{:>02X}".format(r, g, b)
	return color


def get_distinctive_palette(number: int = None) -> List[str]:
	""" A set of selected colors meant to be as clearly distinct from each other as possible."""
	if number:
		return distinctive_palette + [random_color() for _ in range(number)]
	return distinctive_palette



from typing import *

import pandas

from . import colorset
from .palette_annotation import generate_annotation_palette
from .palette_distinctive import generate_distinctive_palette
from .palette_lineage import generate_lineage_palette

def generate_palette(edges: pandas.Series, custom_palette: Dict[str, str] = None, annotations: Dict[str, List[str]] = None,
		kind = 'distinctive') -> Dict[str, str]:
	"""
	Parameters
	----------
	edges: A list of unique genotypes or the edges table.
	custom_palette: Dict[str,str]
	annotations
	kind
	"""
	if custom_palette is None: custom_palette = {}
	if annotations is None: custom_palette = {}

	unique_genotypes = list(edges.index)

	if kind == 'lineage':
		palette = generate_lineage_palette(edges)
	elif kind == 'annotation':
		palette = generate_annotation_palette(annotations, custom_palette)
	else:
		palette = generate_distinctive_palette(unique_genotypes)
	for label in unique_genotypes:
		new_color = custom_palette.get(label, palette.get(label))
		if new_color is None:
			palette[label] = colorset.random_color()

	palette['genotype-0'] = "#FFFFFF"
	palette['removed'] = "#333333"
	return palette



from pathlib import Path


def mullerplot(edges: Path, population: Path, output: Path):
	table_edges = dataio.import_table(edges)
	table_population = dataio.import_table(population)
	from muller.graphics import MullerPlot
	muller_formatter = dataio.GenerateMullerDataFrame()
	diagram_generator = MullerPlot(outlines = True, render = True)
	muller_df = muller_formatter.run(table_edges, table_population)
	diagram_generator.plot(muller_df, output)


def timeseriesplot(filename:Path):
	""" Plots a table of trajectories/genotypes"""
	table = dataio.import_table(filename)
	if 'Trajectory' in table.columns:
		table = table.set_index('Trajectory')
	else:
		table = table.set_index('Genotype')



def benchmark(dataset: Path, output: Path):
	from muller.clustering.metrics.distance_calculator import benchmark, plot_benchmark_results
	dataset = dataio.import_table(dataset)
	benchmark_results = benchmark(dataset)
	plot_benchmark_results(benchmark_results, output)


def main(arguments)->None:
	from muller.workflows.workflow_full import run_workflow
	muller_workflow = run_workflow(arguments)
if __name__ == "__main__":
	import sys
	from pathlib import Path

	sys.path.append(str(Path(__file__).parent.parent))
	from muller import dataio, commandline_parser

	program_parser = commandline_parser.create_parser()
	# Custom method to select `lineage` as the default parser. Used to keep the current api, but will probably be changed later.
	#args = program_parser.parse_args()
	program_arguments = commandline_parser.get_arguments()

	# Need to make sure the argument defaults are applied.

	if program_arguments.name is None:
		program_arguments.name = "lineage"

	if program_arguments.name == 'lineage':
		main(program_arguments)
	else:
		print("Need to use `muller/ lineage --input [input]")





import matplotlib.pyplot as plt
import pandas
from loguru import logger

from muller import widgets


class DerivativePlot:
	def __init__(self):
		pass

	# noinspection PyMethodMayBeStatic
	def run(self, left: pandas.Series, right: pandas.Series):
		detected_left, detected_right = widgets.get_valid_points(left, right, 0.03, 0.97, inner = True)

		diff_left = detected_left.diff()
		diff_right = detected_right.diff()
		diff_series = list(zip(diff_left, diff_right))
		fig, ax = plt.subplots(figsize = (15, 15))
		ax.scatter(diff_left.values, diff_right.values)

		ax.set_xlabel("left derivative")
		ax.set_ylabel("right derivative")

		ax.set_xlim(-1, 1)
		ax.set_ylim(-1, 1)

		ax.axhline(0)
		ax.axvline(0)
		ax.plot([-1, 1], [-1, 1])
		ax.plot([-1, 1], [1, -1])
		correlated = sum([distance_correlated(i) for i in diff_series[1:]])
		anticorrelated = sum([distance_anticorrelated(i) for i in diff_series[1:]])

		logger.info(f"Correlation score: {correlated}")
		logger.info(f"Anticorrelation score: {anticorrelated}")

		plt.show()


import math


def distance_correlated(point) -> float:
	x, y = point
	d = abs(x - y) / math.sqrt(2)
	return d


def distance_anticorrelated(point):
	x, y = point
	d = abs(x + y) / math.sqrt(2)
	return d



from typing import *

import pandas
from shapely import geometry

Number = Union[float, int]
PointType = Tuple[Number, Number]
import collections
import itertools

# There is an issue when attempting to calculate the area of a series. Rather than trying to calculate on individual polygons,
# We use a very small value rather than 0 so that the entire series is technically a single polygon.
# shapely throws a TopologyError if this isn't done anytime you try to get the subset of two polygons.
# The minimum value shouldn't change the calculated area enough to affect anything.
MINIMUM = 0.0001


def sliding_window_legacy(iterable: List, width: int):
	for index in range(len(iterable)):
		try:
			if index + width > len(iterable):
				raise StopIteration  # Because otherwise will still return last last n elements.
			yield tuple(iterable[index:index + width])
		except (IndexError, StopIteration):
			break


def sliding_window(seq, n):
	""" A sequence of overlapping subsequences
	>>> list(sliding_window(2, [1, 2, 3, 4]))
	[(1, 2), (2, 3), (3, 4)]
	This function creates a sliding window suitable for transformations like
	sliding means / smoothing
	>>> mean = lambda seq: float(sum(seq)) / len(seq)
	>>> list(map(mean, sliding_window(2, [1, 2, 3, 4])))
	[1.5, 2.5, 3.5]
	"""
	return zip(*(collections.deque(itertools.islice(it, i), 0) or it
		for i, it in enumerate(itertools.tee(seq, n))))


def get_neighbor(index: pandas.Index, item: Any, previous: bool = True) -> Any:
	"""
		Return the neighboring value to `item`.
	Parameters
	----------
	index: The full index of the series. NOT just the index of detected values
	item
	previous

	Returns
	-------

	"""
	# Get the integer position of the item
	index = list(index)  # pandas.Index uses circular indexing, so the last key is treated as being prior to the initial key.
	offset = -1 if previous else 1

	location = index.index(item)
	if location != 0:
		location += offset
	try:
		return index[location]
	except IndexError:
		return item


def _decompose_correct_split_series(series: pandas.Series):
	""" Corrects the case where a series is undetected at a single point, but otherwise is detected.
		This will cause a topology error in shapely.
	exmple:
		[0.97, 0.0, 0.97, 0.97, 0.97, 0.87, 0.97]
	"""
	default_value = 0.001  # Should be small enough to not affect much.
	# Figure out which points need to be modified.
	indicies = list()

	for index, window in enumerate(sliding_window(series, 3)):
		if window[0] > 0 and window[2] > 0 and window[1] == 0:
			indicies.append(index)
	for i in indicies:
		# guaranteed to not be the last index due to how sliding_window() works.
		series[i + 1] = default_value
	return series


def _correct_one_dimensional_sections(points: List[Tuple[float, float]]) -> List[Tuple[float, float]]:
	""" Removes adjacent 0's representing a one-dimensional offshoot form the polygon."""

	for index, (x, y) in enumerate(points):
		if y == 0:
			try:
				xp, yp = points[index - 1]
				if yp == 0:
					points[index - 1] = (xp, 0.001)
			except IndexError:
				pass
	return points



def get_vertices(series: Union[List[PointType], pandas.Series], use_index:bool = False) -> List[PointType]:
	""" Converts a trajectory/genotype timeseries into a 2D polygon"""
	# Make sure one-dimensional points are omitted.
	# Use a very small value rather than `0` due to how shapely calculates area.
	#minimum = 0.0001
	# Note: Apparently the area is properly calculated now as it was intended to, so this hack isn't necessary anymore.
	# Will keep the process but use `0` as the minimum value in case the hack needs to be re-enabled.
	if not isinstance(series, pandas.Series):
		series = pandas.Series(series)
	masked_series = series.mask(lambda s: s < MINIMUM, MINIMUM).tolist()
	masked_series[0] = series.values[0]
	masked_series[-1] = series.values[-1]

	if use_index:
		# This will use the actual timepoint values rather than a linear range of identically spaced values.
		x_values = series.index
	else:
		x_values = list(range(len(series)))

	points = list(zip(x_values, masked_series))
	if points: # Check if the list is empty
		first_point = points[0]
		if first_point[1] != 0:
			points = [(first_point[0], 0)] + points

		last_point = points[-1]
		if last_point[1] != 0:
			points.append((last_point[0], 0))
	return points


def separate_main(series: List[PointType]) -> List[List[PointType]]:
	""" Separates a series into segregated polygons assuming there is a region of no detection between them."""
	# Assume each series has at least three points.
	if len(series) < 3:
		# Does nothave enough vertices.
		return []
	if not is_multiple(series):
		return [series]
	for index, element in enumerate(sliding_window(series, 2), start = 1):  # Start indexing at 1 since starting with second element.
		if element[0][1] == 0 and element[1][1] == 0:
			# Separate the polygons to the left and right of this region.
			left = series[:index]
			right = series[index:]
			return separate_main(left) + separate_main(right)


def isolate_polygons(series: List[PointType]) -> List[List[PointType]]:
	""" Implemented here until I refactor the above function to omit small sequences"""
	return [i for i in separate_main(series) if len(i) > 2]


def is_multiple(series: List[Tuple[Number, Number]]):
	# Test if one of the series has regions where it is undetected between other areas where it is detected.
	# Basically whether the series needs to be represented by two separate points
	for window in sliding_window(series, 2):
		if window[0][1] == 0 and window[1][1] == 0:
			return True
	return False


def as_polygon(series: pandas.Series) -> geometry.MultiPolygon:
	""" Converts a series object to a polygon object.
		[0.97 0.   0.97 0.97 0.97 0.87 0.97] Fails
	"""
	points = get_vertices(series)
	result = geometry.Polygon(points)
	return result



from .dendrogram import plot_dendrogram
from .lineageplot import flowchart
from .heatmap import plot_heatmap
from .muller_plot import MullerPlot
from .panels import MullerPanel, TimeseriesPanel, generate_lineage_panel
from .plottimeseries import TimeseriesPlot
from .distancegraphs import generate_distance_plot
from . import palettes
Palette = palettes.Palette




import math

import pandas
from loguru import logger
from shapely import geometry
from shapely.errors import TopologicalError
from typing import *
Number = Union[float, int]

from . import polygon


DEBUG = False

def calculate_area(series:pandas.Series)->float:
	values = list()

	for previous, current in zip(series.values[:-1], series.values[1:]):
		if previous < current:
			lower = previous
			higher = current
		else:
			lower = current
			higher = previous

		area = lower + ((higher - lower)/ 2)
		values.append(area)
	return sum(values)



def area_of_series(series: pandas.Series) -> float:
	"""
		Calculates the area under the curve of a given series. Assumes that the series
		can be considered a sequence of polygons where vertices correspond to the
		x-values and y-values in the series.
	"""
	points = polygon.get_vertices(series)
	collection_of_polygons:List[List[Tuple[float,float]]] = polygon.isolate_polygons(points)
	area = 0
	for poly in collection_of_polygons:
		element_area = geometry.Polygon(poly).area
		area += element_area
	return area


def calculate_common_area(left: pandas.Series, right: pandas.Series) -> float:
	""" Calculates |X \cap Y|"""
	# Use the pandas method for now. It is probably slower, but shouldn't add too much time, and should be more readable.

	left_points = polygon.get_vertices(left) if isinstance(left, pandas.Series) else left  # Only decompose if the input was not a list of points.
	right_points = polygon.get_vertices(right) if isinstance(right, pandas.Series) else right

	# Test if one of the series has regions where it is undetected between other areas where it is detected.
	# Basically whether the series needs to be represented by two separate points
	# Since polygon.separate() can handle the case where the series does not need to be modified, may as well run it on all
	# series instead of testing for disjoint polygons first.
	left_collection = polygon.isolate_polygons(left_points)
	right_collection = polygon.isolate_polygons(right_points)

	left_collection_polygon = [geometry.Polygon(i) for i in left_collection if len(i) > 2]
	right_collection_polygon = [geometry.Polygon(i) for i in right_collection if len(i) > 2]

	left_collection_polygon = geometry.MultiPolygon(left_collection_polygon)
	right_collection_polygon = geometry.MultiPolygon(right_collection_polygon)
	try:
		intersection = left_collection_polygon.intersection(right_collection_polygon)
	except TopologicalError as exception:
		logger.error(f"Cannot calculate the intersection:")
		logger.error(f"\t{left_collection_polygon}")
		logger.error(f"\t{right_collection_polygon}")
		raise exception

	total_intersection_area = intersection.area

	return total_intersection_area


def X_and_Y_polygon(left: geometry.MultiPolygon, right: geometry.MultiPolygon) -> float:
	try:
		return left.intersection(right).area
	except TopologicalError:
		logger.error(f"Polygonal error")
		return 0


def X_or_Y_numeric(left: float, right: float, x_and_y: float) -> float:
	""" Calculates the OR area between `left` and `right`
		Parameters
		----------
		left, right: pandas.Series
			The two series to calculate the exclusive area on.
		x_and_y: float
			Precomputed values for the area of `left`, `right`, and x_and_y.
	"""
	# Need to subtract `x_and_y` since it is counted twice in `left` and `right`
	return left + right - x_and_y


def X_or_Y(left: pandas.Series, right: pandas.Series) -> float:
	""" Calculates the OR area between `left` and `right`
		Parameters
		----------
		left, right: pandas.Series
			The two series to calculate the exclusive area on.
	"""

	left_polygons = polygon.as_polygon(left)
	right_polygons = polygon.as_polygon(right)

	multipolygon = geometry.MultiPolygon([left_polygons, right_polygons])
	area = multipolygon.area - left_polygons.intersection(right_polygons).area

	return area


def X_xor_Y(left: pandas.Series, right: pandas.Series) -> float:
	left_polygons = polygon.as_polygon(left)
	right_polygons = polygon.as_polygon(right)

	xor_polygons = left_polygons.symmetric_difference(right_polygons)
	return xor_polygons.area


def difference_polygon(left: geometry.MultiPolygon, right: geometry.MultiPolygon) -> float:
	""" Returns the area of `left` not in `right`"""
	try:
		area = left.difference(right).area
	except TopologicalError:
		logger.error(f"Polygonal error")
		area = 0
	return area


def jaccard_distance_numeric(x_or_y: float, x_and_y: float) -> float:
	""" Calculated the jaccard distance between two series using pre-computed area values.
		Parameters
		----------
		x_or_y, x_and_y: float
			precomputed x_or_y and x_and_y, respectively.
	"""
	return (x_or_y - x_and_y) / x_or_y


def jaccard_distance(left: pandas.Series, right: pandas.Series) -> float:
	""" Calulated the jaccard distance between two series.
		Parameters
		----------
		left, right: pandas.Series
			The two genotypes to calculate the jaccard distance on.
	"""

	x_or_y = X_or_Y(left, right)
	x_and_y = X_and_Y(left, right)
	return jaccard_distance_numeric(x_or_y, x_and_y)


def jaccard_subset_numeric(left: float, right: float) -> float:
	""" Operates on pre-computed areas to help reduce redundant calculations.
		Calculates the jaccard distance assuming `right` is a subset of `left`.
	Parameters
	----------
	left, right: pandas.Series
	left, right: float
	"""
	return (left - right) / left


def jaccard_subset(left: pandas.Series, right: pandas.Series) -> float:
	"""
		Calculates the jaccard distance assuming `right` is a subset of `left`.
	Parameters
	----------
	left, right: pandas.Series
	left, right: float
	"""
	left_area = area_of_series(left)
	right_area = area_of_series(right)
	return jaccard_subset_numeric(left_area, right_area)


def is_subset(left: pandas.Series, right: pandas.Series) -> bool:
	""" A rewrite that uses the shapely library instead."""
	left_poly = polygon.as_polygon(left)
	right_poly = polygon.as_polygon(right)

	try:
		area_intersection = left_poly.intersection(right_poly).area
	except TopologicalError as exception:
		logger.error(f"{left.values}, {left.index}")
		logger.error(f"{right.values}, {right.index}")
		logger.error(f"Left polygon: {left_poly}")
		logger.error(f"Right polygon: {right_poly}")
		logger.error(f"TopologyError")
		return False

	area_right = right_poly.area

	result = math.isclose(area_intersection, area_right, abs_tol = 0.03 ** 2)  # Tolerance is square of the detection limit.
	return result


def is_subset_polygon(left: geometry.MultiPolygon, right: geometry.MultiPolygon) -> bool:
	area_union = left.union(right).area
	area_intersection = left.intersection(right).area
	area_left = left.area
	area_right = right.area

	result = math.isclose(area_intersection, area_right, abs_tol = 0.03 ** 2)

	jaccard_expected = (area_left - area_right) / area_left
	jaccard_actual = (area_union - area_intersection) / area_union

	if DEBUG:
		logger.debug(f"left, right -> {area_left}, {area_right}")
		logger.debug(f"Je = ({area_left:.2f} - {area_right:.2f}) / {area_left:.2f} = {jaccard_expected:.2f}")
		logger.debug(f"Ja = ({area_union:.2f} - {area_intersection:.2f}) / {area_union:.2f} = {jaccard_actual:.2f}")

		logger.debug(f"{jaccard_expected} == {jaccard_actual} -> {result}")
	result = math.isclose(jaccard_expected, jaccard_actual, abs_tol = 0.1)
	return result


X_and_Y = calculate_common_area






import math
from functools import partial
from typing import Any, Iterable, List, Optional, Union

import pandas

Numeric = Union[int, float]

def lag_gens(x: int, values: Iterable[int]) -> int:
	# function to get the generation previous to a specified generation:
	# Implemented in-line so that we can use the population['Genotypes'] within the method.
	candidates = [i for i in values if i < x]
	if candidates:
		return max(candidates)
	else:
		return 0


def dup(iterable: List[Any]) -> List[bool]:
	""" returns a list indicating whether the coresponding index in `iterable` is a duplicated value."""
	return [iterable.count(i) > 1 for i in iterable]


class Vector:
	""" Emulates the way R values can be added to a scalar variable."""

	def __init__(self, value: Optional[Any]):
		self.data: List[Any] = list()
		if value:
			self.data.append(value)

	def __len__(self)->int:
		return len(self.data)

	def __setitem__(self, key: int, value: Any):
		# Inserts `value` at the position `key`
		if key < len(self):
			self.data[key] = value
		elif key - len(self) <= 1:
			self.data.append(value)
		else:
			message = f"Trying to access a value that is not the next position in the array: {key} -> {len(self)}"
			raise ValueError(message)

	def __getitem__(self, item)->Any:
		return self.data[item]


class AdjacencyMatrix:
	def __init__(self, edges: pandas.DataFrame, ages: Optional[pandas.Series] = None):
		self.table_edges = edges

		if ages is not None:
			self.ages = ages
			self.clean_edges = self.generate_clean_edges(ages)
		else:
			self.ages = None
			self.clean_edges = None

	def _validate_point(self, value: int):
		if value not in self.clean_edges['Identity'].values and value not in self.clean_edges['Parent'].values:
			message = f"This is an invalid value: {value}."
			raise ValueError(message)

	def generate_clean_edges(self, lookup: pandas.Series) -> pandas.DataFrame:
		""" This is a more concise implementation of the method used in ggmuller.
			'Replace each genotype name in adjacency matrix with corresponding Age'
		"""
		intermediate_edges = self.table_edges.copy(deep = True)  # To avoid unintented changes.
		intermediate_edges['Identity'] = [lookup.loc[i] for i in self.table_edges['Identity'].values]
		intermediate_edges['Parent'] = [lookup.loc[i] for i in self.table_edges['Parent'].values]
		edges = intermediate_edges.sort_values(by = "Identity")
		self.clean_edges = edges
		self.ages = lookup
		return edges

	def find_start_node(self):
		# Attempts to find the first node in the matrix
		start = self.clean_edges['Parent'].sort_values().iloc[0]  # Just a guess

		while True:
			value = self.move_up(start)
			if value == start: return value
			else:
				start = value

	def move_up(self, identity: int) -> int:
		"""
		Move to parent in adjacency matrix

		Returns the corresponding Parent value.
		When there is no parent (i.e. at the top of the tree), returns the input Identity.
		"""
		# Check if the identity is valid
		self._validate_point(identity)

		parent = self.clean_edges[self.clean_edges["Identity"] == identity].squeeze()  # To make sure it is a series rather than a dataframe
		if parent.empty:
			return identity
		else:
			return parent['Parent']

	def move_down(self, parent: int) -> int:
		self._validate_point(parent)
		daughters: pandas.Series = self.clean_edges[self.clean_edges['Parent'] == parent]['Identity']
		if daughters.empty:
			return parent
		else:
			return daughters.min()

	def move_right(self, identity: int) -> int:
		self._validate_point(identity)

		parent = self.clean_edges[self.clean_edges['Identity'] == identity].squeeze()
		if parent.empty:
			return identity
		parent = parent['Parent']

		siblings = self.clean_edges[self.clean_edges['Parent'] == parent]['Identity'].sort_values().tolist()
		# Find the index of the next value after the one given as the identity.
		try:
			sibling = siblings[siblings.index(identity) + 1]
			return sibling
		except IndexError:
			return identity

	def path_vector(self) -> List[str]:
		n = 0  # Arrays start at zero in python.
		path = self.find_start_node()
		path = Vector(path)  # Since variables in R behave both as scalar and vector values.

		upped = False

		for i in range(10000):  # Prevents endless loops. change to a while loop later
			if not upped:
				for j in range(1000):
					n = n + 1
					path[n] = self.move_down(path[n - 1])
					upped = False
					if path[n] == path[n - 1]: break
			if self.move_right(path[n]) != path[n]:
				n = n + 1
				path[n] = self.move_right(path[n - 1])
				upped = False
			elif self.move_up(path[n]) != path[n]:
				n = n + 1
				path[n] = self.move_up(path[n - 1])
				upped = True

			if path[n] == path[0]: break
			if n > 2 * len(self.clean_edges) + 2:
				message = "Error: stuck in a loop"
				raise ValueError(message)
		# TODO add the final check. Not included yet to save time.from
		if len(path) != (2 * len(self.clean_edges) + 2):
			message = "Error: the adjacency matrix seems to be bipartite"
			raise ValueError(message)

		# Map the path of ages to the corresponding genotype
		path = [self.age_to_genotype(i) for i in path]

		return path

	def age_to_genotype(self, age: int) -> str:
		""" Returns the genotype represented by the given age."""
		if self.ages is None:
			message = f"The lookup table has not been provided yet."
			raise ValueError(message)
		counts = self.ages.value_counts()
		if any(i > 1 for i in counts.values):
			message = "Two or more genotypes were assigned the same age."
			raise ValueError(message)

		reverse_lookup = {v: k for k, v in self.ages.items()}

		return reverse_lookup[age]


class GenerateMullerDataFrame:
	""" implements the get_Muller_df function from ggmuller. The original fails when there are a large number of sampled timepoints."""

	def __init__(self):
		self.time_column = 'Generation'
		self.identity_column = 'Identity'
		self.population_column = 'Population'

		# Initial population size. Will probably never be changed.
		self.initial_population_size = 0
		self.initial_start_positions = 0.5
		self.detection_limit = 0  # Would normally use 0.05 * 100, but the original script uses 0.



	# Add the sub-workflows that will be used.
	@staticmethod
	def _adjust_population_table(population: pandas.DataFrame, first_generation: pandas.DataFrame, start_positions: float) -> pandas.DataFrame:
		# copy all rows for generations at which new genotypes appear
		new_rows = population[population['Generation'].isin(first_generation['start_time'].values)]
		prev_rows = population[population['Generation'].isin(first_generation['previous_time'].values)]
		prev_rows.index = [i + 1 for i in prev_rows.index]

		# adjust generations of copied rows
		partial_lag_gens = partial(lag_gens, values = population['Generation'].unique())
		adjusted_generation:List[float] = new_rows['Generation'] - start_positions * (new_rows['Generation'] - new_rows['Generation'].apply(partial_lag_gens))

		adjusted_population = ((1 - start_positions) * new_rows['Population']) + (start_positions * prev_rows['Population'])

		new_rows['Generation'] = adjusted_generation
		new_rows['Population'] = adjusted_population

		return new_rows.reset_index(drop = True) # To make the tale more consistent

	@staticmethod
	def _clean_population_table(population: pandas.DataFrame) -> pandas.DataFrame:
		""" Implements the cleaning steps from the ggmuller package. Since these are already done with the other scripts, this returns
			The input table.
		"""
		return population


	def _find_start_positions(self, series: pandas.Series, start_positions: float) -> float:
		all_generations = series.unique()
		difference_time_diff = series.diff().abs().min()  # .diff() is reversed from r, so use abs to correct it.
		difference_time_absolute = max(all_generations) - min(all_generations)
		delta = min(1E-2 * difference_time_diff, 1E-4 * difference_time_absolute)

		start_positions = max(start_positions, delta)
		start_positions = min(start_positions, 1 - delta)

		# I think the ggmuller scripts assumed there would be relatively small time intervals, such as `35 days` or `120` minutes
		# The LTEE has several thousand, so the above code results in a negative value. For now replace negative values with the original
		# values of `start_positions`.
		start_positions = max(start_positions, self.initial_start_positions)

		return start_positions

	def _get_initial_generations(self, population: pandas.DataFrame) -> pandas.DataFrame:
		""" Maps each genotype to both the first timepoint it appears as well as the previous timepoint."""
		genotype_groups = population.groupby(by = "Identity")
		table = list()
		for identity, group in genotype_groups:
			detected = group[group[self.population_column] > self.detection_limit]
			detected_timepoint = detected[self.time_column].min()

			# Also need to find the previous timepoint.

			previous_timepoints = [i for i in group[self.time_column].values if i < detected_timepoint]
			if previous_timepoints:  # Check if there are any previous timepoints.
				previous_timepoint = max(previous_timepoints)
			else:
				previous_timepoint = detected_timepoint  # Not sure if this is the way the original script handled this edge case.

			row = {
				self.identity_column: identity,
				'start_time':         detected_timepoint,
				'previous_time':      previous_timepoint
			}
			table.append(row)

		df = pandas.DataFrame(table)

		df = df[[self.identity_column, 'start_time', 'previous_time']]  # sort columns to match ggmuller

		# Remove `generation-0` to match the ggmuller script
		df = df[df['start_time'] != 0]
		return df.reset_index(drop = True) # The index isn't needed, so reset it to make it more consistent with the other tables.

	@staticmethod
	def expand(left_values: Iterable[Any], right_values: Iterable[Any]) -> pandas.DataFrame:
		""" Generates a two-column table which pairs every left value with every right value.
			The resulting table will be len(left) * len(right) rows.
		"""
		added_rows = list()
		for left in left_values:
			for right in right_values:
				added_rows.append({'Left': left, 'Right': right})
		added_rows = pandas.DataFrame(added_rows)

		return added_rows

	@staticmethod
	def add_semi_frequencies(population: pandas.DataFrame) -> pandas.DataFrame:
		frequencies = population.groupby(by = "Generation").apply(lambda s: .5 * (s['Population'] / s['Population'].sum()))
		population['Frequency'] = frequencies.fillna(0).values
		population['Population'] = population['Population'] / 2  # Because of the duplication

		return population

	def apply_add_start_points(self, population: pandas.DataFrame, start_positions: float = 0.5) -> pandas.DataFrame:
		start_positions = self._find_start_positions(population['Generation'], start_positions)
		# set small initial population size
		first_generation = self._get_initial_generations(population)

		# if all genotypes appear at the first time point then don't make any changes:
		if len(first_generation['start_time'].unique()) == 1:
			return population

		# Adjust the generations and populations of population
		new_rows = self._adjust_population_table(population, first_generation, start_positions)

		# Add the new columns to the original population dataframe
		pop_df = pandas.concat([population, new_rows])
		pop_df = pop_df.sort_values(by = ["Generation", "Identity"]).reset_index(drop = True)

		# adjust generations in reference list
		first_generation['Generation'] = first_generation["start_time"] - start_positions * (
				first_generation['start_time'] - first_generation['previous_time'])
		# Set small initial populations in reference list
		first_generation['Population2'] = self.initial_population_size

		# replace initial populations in the dataframe with values from reference list
		pop_df = pop_df.merge(first_generation, how = 'outer')
		# Replace the population of rows where `Population2` is defined.
		pop_df['Population'] = [(i if math.isnan(j) else j) for i, j in zip(pop_df['Population'].values, pop_df['Population2'].values)]

		return pop_df[['Generation', 'Identity', 'Population']]

	def get_genotype_ages(self, table: pandas.DataFrame) -> pandas.Series:
		# Apply the filtering step before the groupby step.

		df = table[(table['Population'] > self.detection_limit) | (table['Generation'] == table['Generation'].max())]

		groups = df.groupby(by = 'Identity')

		first = groups.first()
		df = first.sort_values(by = ["Generation", "Identity"], ascending = [True, True])
		df = df.reset_index()
		df['Age'] = list(range(1, len(df) + 1))
		series = df.set_index("Identity")['Age']
		return series

	def reorder_by_vector(self, muller_df: pandas.DataFrame, path_vector: List[str]):
		""" Orders the muller dataframe so that each series is plotted in the correct order."""

		# Add a unique id column to the vector.
		# Identify duplicates
		# Using pandsa.Series.value_counts() might be easier, but would require makeing a Series object.

		vector_df = pandas.DataFrame({
			"vector":           path_vector,
			'uniqueGenotypeId': self._reorder_by_vector_generate_unique_ids_genotype(path_vector)
		})

		# Now start modifying `muller_df`
		# Take the vector with the unique genotype id's and multiply it so that it is the same length as `muller_df`
		number_of_unique_timepoints = len(muller_df['Generation'].unique())
		vector: List[str] = vector_df['uniqueGenotypeId'].tolist() * number_of_unique_timepoints
		# Assume that the `muller_df` is already sorted by generation.
		# Basically, each repeated sequence in `vector` corresponds to a single generation.
		unique_id_generation = ["{}_{}".format(i, g if g != int(g) else int(g)) for i, g in zip(vector, muller_df['Generation'].tolist())]

		# Index the dataframe by the unique generation id so that .loc[] can be used to reorder it.

		# Name it `Unique_id` to match the ggmuller scripts.
		muller_df['Unique_id'] = self._reorder_by_vector_generate_unique_ids_generation(muller_df)
		# To match the scripts
		muller_df['Group_id'] = muller_df['Unique_id'].apply(lambda s: s.split('_')[0])
		muller_df_indexed = muller_df.set_index('Unique_id')
		muller_df_sorted = muller_df_indexed.loc[unique_id_generation]

		return muller_df_sorted.reset_index()

	@staticmethod
	def _reorder_by_vector_generate_unique_ids_generation(df: pandas.DataFrame) -> List[str]:
		seen = set()
		uids = list()
		for index, row in df.iterrows():
			name = row['Identity']
			generation = row['Generation']
			# Check if the `generation` value can be safely converted to int to match ggmuller.
			if generation - int(generation) == 0:
				generation = int(generation)
			key = (name, generation)
			if key in seen:
				name += "a"
			else:
				seen.add(key)
			uid = f"{name}_{generation}"
			uids.append(uid)
		return uids

	@staticmethod
	def _reorder_by_vector_generate_unique_ids_genotype(iterable: List[str]) -> List[str]:
		"""
		Since each series has been split in two, there should be exactly one duplicate for each value.
		Since each value needs to have a unique id, append a 'a' to the second value.
		Here is a quick and dirty implementation:
		"""
		seen = set()
		unique_ids = list()
		for element in iterable:
			if element in seen:
				unique_ids.append(element + 'a')
			else:
				unique_ids.append(element)
				seen.add(element)
		return unique_ids

	def correct_population_values(self, population: pandas.DataFrame) -> pandas.DataFrame:
		""" Adjusts the values in the population dataframe to make sure it can be plotted correctly."""
		# add rows to population to ensure genotype starting positions are plotted correctly.
		population = self.apply_add_start_points(population, self.initial_start_positions)
		# Construct a dataframe with the "Age" of each genotype.
		population_sorted = population.sort_values(by = ["Generation", "Population"], ascending = [True, False])
		# Add semi-frequencies
		frequencies = self.add_semi_frequencies(population_sorted)

		# duplicate the rows on the dataframe since the muller plots need to plot the top and bottom halves of each series individually.
		muller_df = pandas.concat([frequencies, frequencies]).sort_values(by = 'Generation')

		return muller_df

	def run(self, edges: pandas.Series, population: pandas.DataFrame) -> pandas.DataFrame:
		# Test if the length of the population table makes sense given the population and number of timepoints.
		edges = edges.reset_index()
		population = self._clean_population_table(population)
		# Construct a dataframe with the "Age" of each genotype.
		population_sorted = population.sort_values(by = ["Generation", "Population"], ascending = [True, False])

		# Generate a dataframe with the adjusted generations and populations.
		muller_df = self.correct_population_values(population)

		ages = self.get_genotype_ages(population_sorted)

		clean_edges = AdjacencyMatrix(edges, ages)
		path_vector: List[str] = clean_edges.path_vector()
		path_vector: List[str] = path_vector[::-1]  # Convert to list and reverse it to match ggmuller scripts.
		muller_ordered = self.reorder_by_vector(muller_df, path_vector)
		# Replace each age in the path vector with the corresponding genotype name
		# Rearrange the columns so pandas.testing.assert_frame_equal doesn't complain.
		muller_ordered = muller_ordered[['Generation', 'Identity', 'Population', 'Frequency', 'Group_id', 'Unique_id']]
		return muller_ordered



from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
from scipy.cluster import hierarchy

try:
	from muller.clustering.metrics import DistanceCache
except ModuleNotFoundError:
	from ..clustering.metrics import DistanceCache


def plot_dendrogram(linkage_table: Any, labels, filename: Path):
	# noinspection PyUnusedLocal
	fig, ax = plt.subplots(figsize = (15, 15))
	ax: plt.Axes
	linkage_table = linkage_table[['left', 'right', 'distance', 'observations']].values # Removes extra column
	# plt.figure(figsize = (15, 15))
	ax.set_title('Hierarchical Clustering Dendrogram', size = 40)
	ax.set_xlabel('Trajectory Label', size = 32)
	ax.set_ylabel('Distance', size = 32)
	hierarchy.dendrogram(
		linkage_table,
		leaf_rotation = 90,  # rotates the x axis labels
		leaf_font_size = 8,  # font size for the x axis labels,
		labels = labels,
		ax = ax
	)
	ax.tick_params(axis = 'both', labelsize = 20)

	plt.savefig(filename, dpi = 500)



import io
from pathlib import Path
from typing import Optional, Union

import pandas
from loguru import logger
from muller import widgets

# noinspection PyProtectedMember
def _import_table_from_path(filename: Path, sheet_name: Optional[str] = None, index: Optional[str] = None) -> pandas.DataFrame:
	""" Imports a file as a pandas.DataFrame. Infers filetype from the filename extension/suffix.
	"""
	if filename.suffix in {'.xls', '.xlsx'}:
		data: pandas.DataFrame = pandas.read_excel(str(filename), sheet_name = sheet_name)
	else:
		sep = '\t' if filename.suffix in {'.tsv', '.tab'} else ','
		try:
			data: pandas.DataFrame = pandas.read_csv(str(filename), sep = sep)
		except UnicodeDecodeError:
			from bs4 import UnicodeDammit
			from io import StringIO
			contents = filename.read_bytes()
			unicode_contents = UnicodeDammit(contents)
			data: pandas.DataFrame = pandas.read_csv(StringIO(unicode_contents.unicode_markup), sep = sep)

	if index and index in data.columns:
		data = data.set_index(index)

	return data


def _import_table_from_string(string: str, delimiter: Optional[str] = None, index: Optional[str] = None) -> pandas.DataFrame:
	""" Imports a table represented as a basic string object."""
	# Remove unwanted whitespace.
	string = '\n'.join(i.strip() for i in string.split('\n') if i)
	if not delimiter:
		delimiter = '\t' if '\t' in string else ','
	result = pandas.read_csv(io.StringIO(string), sep = delimiter, index_col = False)
	if index:
		# Using `index_col` in `read_table()` doesn't work for some reason.
		result[index] = result[index].astype(str)
		result.set_index(index, inplace = True)
	return result


def import_table(input_table: Union[str, Path], sheet_name: Optional[str] = None, index: Optional[str] = None) -> pandas.DataFrame:
	if isinstance(input_table, Path):
		data = _import_table_from_path(input_table, sheet_name, index)
	else:
		data = _import_table_from_string(input_table, index = index)
	# Make sure the x-values are numeric
	numeric_columns = widgets.get_numeric_columns(data.columns)
	nonnumeric_columns = [i for i in data.columns if i not in numeric_columns]

	def _cast_to_int(value)->int:
		try:
			return int(value)
		except (ValueError, TypeError):
			return value

	try:
		# Using float causes problems
		data.columns = [_cast_to_int(i) for i in data.columns]
	except ValueError:
		# The columns could not be converted to str
		logger.warning(f"The columns could not be converted to int: {data.columns}")
	try:
		# Keep str columns to the left
		sorted_columns = sorted(data.columns, key = lambda s: s if isinstance(s, int) else -1)
		data = data[sorted_columns]
	except TypeError as exception:
		logger.error(data.columns)
		raise exception
	return data



import itertools
import math
import multiprocessing
from typing import Dict, List, Optional, Tuple, Generator

import pandas
from loguru import logger

try:
	# tqdm is an optional dependancy
	from tqdm import tqdm
except ModuleNotFoundError:
	tqdm = None

try:
	from muller.clustering.metrics import distance_methods
	from muller import widgets
except ModuleNotFoundError:
	from . import distance_methods
	from ... import widgets


class DistanceCalculator:
	""" Refactor to clean up code and add multithreading."""

	def __init__(self, detection_limit: float, fixed_limit: float, metric: str, threads: Optional[int] = None):
		self.detection_limit = detection_limit
		self.fixed_limit = fixed_limit
		self.metric = metric
		self.threads = threads
		# Basically used as a cache. Should save memory compared to loading each pair of trajectories directly into `pair_combinations`.
		self.trajectories: Optional[pandas.DataFrame] = None

		self.progress_bar_minimum_points = 10000  # The value to activate the scale bar at.

	def calculate_pairwise_distances_threaded(self, pair_combinations: Generator, total:Optional[int] = None)->Dict[Tuple[str,str], float]:
		""" Threaded version of the pairwise distance calculator"""
		pair_array: Dict[Tuple[str,str], float] = dict()
		progress_bar = tqdm(total = total)
		pool = multiprocessing.Pool(processes = self.threads)
		for i in tqdm([pool.apply_async(calculate_distance, args = (self, e, self.trajectories)) for e in pair_combinations]):
			key, value = i.get() # retrieve the calculated value
			pair_array[key] = value
			pair_array[key[::-1]] = value
			progress_bar.update(1)

		return pair_array

	def calculate_pairwise_distances_serial(self, pair_combinations: Generator, total:Optional[int] = None):
		""" Nonthreaded version of the pairwise distance calculator"""
		pair_array: Dict[Tuple[str,str], float] = dict()
		progress_bar = tqdm(total = total)

		for element in pair_combinations:
			key, value = calculate_distance(self, element, self.trajectories)
			pair_array[key] = value
			pair_array[key[1], key[0]] = value  # It's faster to add the reverse key rather than trying trying to get  test forward and reverse keys
			progress_bar.update(1)
		return pair_array


	def calculate_pairwise_distances(self, labels: List[str]) -> Dict[Tuple[str, str], float]:
		""" Implements the actual loop over all pairs of trajectories.
		"""
		# May as well move the combination function here so we don't have to pass an additional parameter specifying the total number
		# of trajectories so tqdm workd properly.

		total_elements = len(labels)
		total_combinations = widgets.calculate_number_of_combinations(total_elements)

		logger.debug(f"Generating pairwise combinations with {total_elements} items resulting in {total_combinations} combinations...")

		if total_combinations > 100_000_000:
			message = f"The provided dataset has {total_elements} trajectories, which requires {total_combinations} distance calculations." \
					  "This will require a long time to process (you may need to adjust the number of available threads with the --threads option)" \
					  "and will consume a large amount of memory (i.e. more than 16GB)."
			logger.warning(message)
		pair_combinations = widgets.get_pair_combinations(labels)

		if self.threads and self.threads > 1:  # One process is slower than using the serial method.
			logger.debug(f"Using multithreading...")
			pair_array = self.calculate_pairwise_distances_threaded(pair_combinations, total_combinations)
		else:
			logger.debug(f"Using a single thread...")
			pair_array = self.calculate_pairwise_distances_serial(pair_combinations, total_combinations)

		# Assume that any pair with NAN values are the maximum possible distance from each other.
		try:
			maximum_distance = max(filter(lambda s: not math.isnan(s), pair_array.values()))
		except ValueError:
			# Usually cause by max() being used on an empty sequence.
			message = f"Could not calculate the pairwise distances due to invalid series (usually because all measurements are below the detectionlimit"
			raise ValueError(message)

		pair_array = {k: (v if not math.isnan(v) else maximum_distance) for k, v in pair_array.items()}

		return pair_array

	def run(self, trajectories: pandas.DataFrame) -> Dict[Tuple[str, str], float]:
		"""
			Calculates the distance between all pairwise combinations of mutational trajectories. The total number of pairs can be calculated by
			`n!/(n-2)!`, which can be simlified to `n*(n-1)`.
			10 trajectories -> 90
			100 trajectories -> 9900
			1000 trajectories -> 999000
			31000 trajectories -> 9.61E8

		Parameters
		----------
		trajectories
		"""
		logger.debug("Calculating the pairwise values...")
		logger.debug(f"\t detection limit: {self.detection_limit}")
		logger.debug(f"\t fixed limit: {self.fixed_limit}")
		logger.debug(f"\t metric: {self.metric}")
		logger.debug(f"\t threads: {self.threads}")

		self.trajectories = trajectories

		pairwise_distances = self.calculate_pairwise_distances(trajectories.index)

		return pairwise_distances


# Keep this as a separate function. Class methods are finicky when used with multiprocessing.
def calculate_distance(process:DistanceCalculator, element: Tuple[str, str], trajectories: pandas.DataFrame) -> Tuple[Tuple[str, str], float]:
	""" Implements the actual calculation for a specific pair of trajectories.
		It should be atomitized so that it works with multithreading.
	"""
	left, right = element
	left_trajectory = trajectories.loc[left]
	right_trajectory = trajectories.loc[right]

	# We only care about the timepoints such that `detection_cutoff` < f < `fixed_cutoff.
	# For now, lets require that both timepoints are detected and not yet fixed.
	# There is an issue related to comparing fixed genotypes against non-fixed genotypes.
	left_was_fixed = widgets.fixed(left_trajectory, process.fixed_limit)
	right_was_fixed = widgets.fixed(right_trajectory, process.fixed_limit)

	if left_was_fixed == right_was_fixed:
		left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, process.detection_limit, process.fixed_limit,
			inner = False)
	else:
		left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, process.detection_limit, inner = False)

	if left_reduced.empty or right_reduced.empty:
		# Treat both trajectories as fixed immediately.
		distance_between_series = fixed_overlap(left_trajectory, right_trajectory, process.fixed_limit)
	else:
		distance_between_series = distance_methods.calculate_distance(left_reduced, right_reduced, process.metric)
	return element, distance_between_series


def fixed_overlap(left: pandas.Series, right: pandas.Series, fixed_cutoff: float) -> float:
	"""
		Calculates the overlap of two series that are both undetected or fixed at all timepoints.
		This will be the distance value for these two series.
	Parameters
	----------
	left:pandas.Series
	right:pandas.Series
	fixed_cutoff: float
	"""
	left_fixed: pandas.Series = left[left.gt(fixed_cutoff)].dropna()
	right_fixed: pandas.Series = right[right.gt(fixed_cutoff)].dropna()
	if left_fixed.empty or right_fixed.empty:
		# Both are undetected, since they both failed the invalid timepoint filter.
		p_value = False
	else:
		# Since these genotypes fixed immediately, they should only be grouped together if they
		# fixed at the same timepoint.
		fixed_at_same_time = left_fixed.index[0] == right_fixed.index[0]
		# Check if the trajectories overlap completely
		overlap = set(left_fixed.index) & set(right_fixed.index)
		complete_overlap = len(overlap) == len(left_fixed) and len(overlap) == len(right_fixed)
		p_value = fixed_at_same_time and complete_overlap
	result = 0 if p_value else math.nan

	return result


def plot_benchmark_results(benchmarks, output_filename):
	import matplotlib.pyplot as plt
	import seaborn
	labels = list()
	values = list()
	for k, v in benchmarks.items():
		labels.append(k)
		values.append(v)
	fig, ax = plt.subplots(figsize = (12, 10))
	seaborn.barplot(x = values, y = labels)

	ax.xlabel('time in seconds', fontsize = 14)
	ax.ylabel('number of processes', fontsize = 14)
	ax.title('Serial vs. Multiprocessing via Parzen-window estimation', fontsize = 18)
	plt.grid()

	plt.savefig(output_filename)


def benchmark(filename):
	import time

	table = pandas.read_excel(filename)
	table = table.set_index('Trajectory')
	table = table[widgets.get_numeric_columns(table.columns)]
	combinations = list(itertools.combinations(table.index, 2))
	benchmarks = dict()
	for index in [None] + list(range(1, 9)):
		logger.info(f"Running iteration {index}")
		distance_calculator = DistanceCalculator(.03, .97, 'binomial', threads = index)
		start = time.time()
		if index is None:
			results = [calculate_distance(distance_calculator, e, table) for e in combinations]
		else:
			pool = multiprocessing.Pool(processes = index)
			results = [pool.apply_async(calculate_distance, args = (distance_calculator, e, table)) for e in combinations]
			results = [i.get() for i in results]
		duration = time.time() - start
		label = str(index)
		benchmarks[label] = duration
	return benchmarks



from typing import Dict

import matplotlib.pyplot as plt

try:
	from muller.graphics import MullerPlot, TimeseriesPlot, palettes, Palette
except (ModuleNotFoundError, ImportError):
	from .muller_plot import MullerPlot
	from .plottimeseries import TimeseriesPlot
	from . import palettes
	from .palettes import Palette

plt.clf()
# plt.switch_backend('agg')
import pandas
from pathlib import Path
from typing import Optional, Tuple, List

from loguru import logger


class TimeseriesPanel:
	""" Plots the genotype frequencies over time with an inset of the original mutaitonal trajectories.
		Parameters
		----------
			render: bool
				Whether to generate additional renders in svg format of the resulting genotype plot.
			legend: bool
				Whether to include a legend in the plot.
	"""

	def __init__(self, render: bool, legend: bool = True):
		self.render = render

		self.filetypes = ['.png']
		if self.render:
			self.filetypes += ['.svg']

		# Parameters concerning the plot legend.
		self.legend = legend
		self.legend_font_properties = {'size': 12}  # The size of the font used to label each series in the legend.
		self.legend_location = 'right'
		self.legend_title = 'Genotypes'

		self.dpi = 250

		# Set the scaling factors.
		# The plots seem to look well up to 20 or so series, but should probably be increased after that.
		# The current values seem to work well with datasets of 10 to 20 genotypes and 15 to 20 timepoints.
		self.scale_max = 3
		self.reference_x = 20
		self.reference_y = 20

	def _get_scaling_factor(self, width_x: int, width_y: int) -> Tuple[int, int]:
		""" Since very large datasets require the plots to scale up, try to infer a decent scaling factor
			So that the resulting plots are actually legible.
		"""
		scale_x = round(width_x / self.reference_x)
		scale_y = round(width_y / self.reference_y)
		# Make sure they aren't 0
		if scale_x == 0: scale_x = 1
		if scale_y == 0: scale_y = 1

		if scale_x > self.scale_max:
			logger.debug(f"Timeseries Plot: Reducing x scale from {scale_x} to {self.scale_max}.")
			scale_x = self.scale_max
		if scale_y > self.scale_max:
			logger.debug(f"Timeseries Plot: Reducing y scale from {scale_y} to {self.scale_max}.")
			scale_y = self.scale_max
		return scale_x, scale_y


	def run(self, timeseries_genotype: pandas.DataFrame, palette: Palette, filename: Path,
			timeseries_trajectory: Optional[pandas.DataFrame] = None):
		"""
			Plots the clustered muller_genotypes.
		Parameters
		----------
		timeseries_genotype: pandas.DataFrame
			A dataframe where columns correspond to timepoints and rows correspond to individual series to plot.
			This is for a generic timeseries as described and isn't meant for a specific table.
		palette: Palette
			Has information related to how the trajectories relate to each genotype.
		filename: Optional[str, Path]
			Path to save the genotype plot. May include multiple filenames.
		timeseries_trajectory: pandas.DataFrame
			The table with each trajectory used to generate the genotypes.
		"""
		palette_genotypes = palette
		palette_trajectories = palette_genotypes.get_trajectory_palette()

		# Need to scale the graph so very large datasets are still easily distinguishable.
		number_of_timepoints = len(timeseries_genotype.columns)
		number_of_series = len(timeseries_genotype)

		# Calculate what the scale should be given the dataset shape. Any dataset smaller than the reference is automatically converted to 1.
		scale_x, scale_y = self._get_scaling_factor(number_of_timepoints, number_of_series)
		# Create the plotting object

		# Set up the plotting area. The genotype plot will cover all three columns while the trajectory plot will be relegated to the
		# Upper left. The legend will occupy the remaining space in the upper right.
		plotter = TimeseriesPlot()
		# Manually define the figure to control the figure size
		figure = plt.Figure(figsize = (12 * scale_x, 10 * scale_y))
		grid = plt.GridSpec(2, 3, wspace = 0.4, hspace = 0.3, figure = figure)

		# Plot all trajectories
		taxes: plt.Axes = plt.subplot(grid[0, 0:2])
		plotter.plot(timeseries_trajectory, palette_trajectories, taxes)
		gaxes: plt.Axes = plt.subplot(grid[1, :])
		# Plot clustered genotypes. Should be same as above, but colored based on genotype cluster.
		# Plot the mean of each cluster
		plotter.plot(timeseries_genotype, palette_genotypes, gaxes)

		if self.legend:
			bbox = self.get_legend_size(number_of_series)
			if bbox:
				plt.legend(
					loc = self.legend_location,
					ncol = 2,
					bbox_to_anchor = bbox,
					prop = self.legend_font_properties,
					title = self.legend_title,
				)
		if filename:
			self.save_figure(filename)

	@staticmethod
	def get_legend_size(number_of_genotypes: int) -> Optional[Tuple[float, float]]:
		if number_of_genotypes < 20:
			bbox = (1, 2)
		elif number_of_genotypes < 30:
			bbox = (1, 1.8)
		else:
			bbox = None

		return bbox

	def save_figure(self, filename: Path):
		""" Saves the diagram in the format specified by the filename's suffix"""
		plt.savefig(filename, dpi = self.dpi)


class MullerPanel:
	""" A panel of graphics describing the lineage of a dataset."""

	def __init__(self, vertical: bool = True, render: bool = True):
		self.vertical = vertical  # The orientation of the plots.
		self.vertical = False
		self.render = render

	def plot(self, timeseries: pandas.DataFrame, muller_df: pandas.DataFrame, palette: Palette = None, annotations: Dict[str, List[str]] = None,
			filename: Optional[Path] = None):
		if self.vertical:
			figure = plt.figure(figsize = (20, 20))
			grid = plt.GridSpec(2, 1)
			ax_timeseries: plt.Axes = figure.add_subplot(grid[0])
			ax_muller: plt.Axes = figure.add_subplot(grid[1], sharex = ax_timeseries)
		else:
			figure = plt.figure(figsize = (20, 10))
			grid = plt.GridSpec(1, 2, figure = figure)
			ax_timeseries: plt.Axes = figure.add_subplot(grid[0])
			ax_muller: plt.Axes = figure.add_subplot(grid[1], sharey = ax_timeseries)

		plotter_timeseries = TimeseriesPlot(render = self.render, style = 'nature')
		plotter_timeseries.set_scale(1)
		plotter_muller = MullerPlot(outlines = True, render = self.render, style = 'nature')
		plotter_muller.set_scale(1)

		plotter_timeseries.plot(timeseries, palette, ax = ax_timeseries)
		plotter_muller.plot(muller_df, color_palette = palette, ax = ax_muller, annotations = annotations)

		if self.vertical:
			ax_timeseries.xaxis.set_visible(False)
		else:
			ax_muller.yaxis.set_visible(False)

		plt.tight_layout()

		if filename:
			plt.savefig(filename)

	def run_from_ggmuller(self, populations: pandas.DataFrame, edges: pandas.DataFrame, palette: Optional[Dict[str, str]] = None,
			basename: Optional[Path] = None):
		""" Generates The panel object using `population` and `edges` tables."""
		from muller.dataio import GenerateMullerDataFrame
		generator = GenerateMullerDataFrame()
		muller_df = generator.run(edges, populations)

		timeseries = convert_population_to_timeseries(populations)
		return self.plot(timeseries, muller_df, palette)


def convert_population_to_timeseries(populations: pandas.DataFrame) -> pandas.DataFrame:
	""" Converts a `population` table into a `timeseries` table. """
	table = populations.pivot(index = 'Identity', columns = 'Generation', values = 'Population')
	table = table[sorted(table.columns)]

	# Need to make sure the values are within [0,1]
	table = table / 100
	return table


def _generate_image_size(size_muller: Tuple[int, int], size_lineage: Tuple[int, int], vertical: bool) -> Tuple[
	Tuple[int, int], Tuple[int, int], Tuple[int, int]]:
	muller_width, muller_height = size_muller
	lineage_width, lineage_height = size_lineage

	if vertical:
		scale = muller_width / lineage_width
		new_lineage_width = round(scale * lineage_width)
		new_lineage_height = round(scale * lineage_height)

		panel_width = new_lineage_width
		panel_height = muller_height + new_lineage_height

		position = (0, muller_height)
	else:
		scale = muller_height / lineage_height
		new_lineage_width = round(scale * lineage_width)
		new_lineage_height = round(scale * lineage_height)

		panel_width = muller_width + new_lineage_width
		panel_height = new_lineage_height
		position = (muller_width, 0)

	return (new_lineage_width, new_lineage_height), (panel_width, panel_height), position


def generate_lineage_panel(filename_muller: Path, filename_lineage: Path, filename: Path, vertical: bool = True):
	""" Combines the muller plot and lineage plots into a single figure."""

	try:
		from PIL import Image
	except ModuleNotFoundError:
		# Cannot combine the two plots.
		return None
	image_muller: Image = Image.open(filename_muller)
	image_lineage: Image = Image.open(filename_lineage)

	new_lineage_size, panel_size, position = _generate_image_size(image_muller.size, image_lineage.size, vertical)
	# Resize the lineageplot
	image_lineage = image_lineage.resize(new_lineage_size)
	# Generate the new figure
	newimage = Image.new(image_muller.mode, panel_size)
	newimage.paste(image_muller, (0, 0))
	newimage.paste(image_lineage, position)

	newimage.save(filename)


if __name__ == "__main__":
	pass
